:- use_module(library(lists)).

bool(0).
bool(1).

sat(Expr) :-
	sat_to_constraints(Expr, Cs, 1),
	solve(Cs).

sat_to_constraints(Expr, Cs, Val) :-
	simple(Expr), !, Cs = [], Val = Expr.
sat_to_constraints(Expr, Cs, Val) :-
	Expr =.. [F|ExprArgs],
	maplist(sat_to_constraints, ExprArgs, ArgsCs, Args),
	ExprV =.. [F|Args], 
	append([[constraint(Val, ExprV)]|ArgsCs], Cs).

% solve(Cs0): repeatedly applies simplify until no change occurs.
% Then it sets the first variable to O or 1 and recurses.
solve([]) :- !.
solve(Cs0) :-
	simplify(Cs0, Cs), !,
	(   Cs == Cs0 
	->  term_variables(Cs0, [Var|_]),
	    bool(Var), solve(Cs0)
	;   solve(Cs)
	).

% simplify(Cs0, Cs): Cs is obtained by possibly reducing each element of Cs0.
% Fails if a reduction detects failure.
simplify([], Cs) :- !, Cs = [].
simplify([C0|Cs0], Cs) :-
	reducible(C0, Test), !,
	Test, simplify(Cs0, Cs).
simplify([C|Cs0], [C|Cs]) :-
	simplify(Cs0, Cs).

% reducible(Constraint, Test): succeeds if the truth value of Constraint
% can be obtained by running Test.
reducible(constraint(X, A /\ B), Test) :-
	(   A == 0 -> Test = (X = 0)
	;   A == 1 -> Test = (X = B)
	;   B == 0 -> Test = (X = 0)
	;   B == 1 -> Test = (X = A)
	;   X == 1 -> Test = (A = 1, B = 1)
	;   B == A -> Test = (X = A)
	).
reducible(constraint(X, \ A), Test) :-
	(   nonvar(X) -> NotX is 1-X, Test = (A = NotX)
	;   nonvar(A) -> NotA is 1-A, Test = (X = NotA)
	;   X == A    -> Test = fail
	).

end_of_file.

% Az albbi nyomkvetsben a numerikus vltozneveket (remlhetleg)
% rtelmes nevekre rtuk t

| ?- trace, sat(\ (\A /\ \B) /\ \A).
% The debugger will first creep -- showing everything (trace)
        1      1 Call: sat(\ (\A/\ \B)/\ \A) ? 
        2      2 Call: sat_to_constraints(\ (\A/\ \B)/\ \A,_3035,1) ? s
        2      2 Exit: sat_to_constraints(\ (\A/\ \B)/\ \A,[constraint(1,AorB/\NA),constraint(AorB,\AnorB),constraint(AnorB,NA2/\NB),constraint(NA2,\A),constraint(NB,\B),constraint(NA,\A)],1) ? 
        3      2 Call: solve([constraint(1,AorB/\NA),constraint(AorB,\AnorB),constraint(AnorB,NA2/\NB),constraint(NA2,\A),constraint(NB,\B),constraint(NA,\A)]) ? 
        4      3 Call: simplify([constraint(1,AorB/\NA),constraint(AorB,\AnorB),constraint(AnorB,NA2/\NB),constraint(NA2,\A),constraint(NB,\B),constraint(NA,\A)],_5457) ? s
        4      3 Exit: simplify([constraint(1,1/\1),constraint(1,\0),constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B),constraint(1,\0)],[constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B)]) ? 
        5      3 Call: solve([constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B)]) ? 
        6      4 Call: simplify([constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B)],_7481) ? s
        6      4 Exit: simplify([constraint(0,1/\NB),constraint(1,\0),constraint(NB,\B)],[constraint(0,1/\NB),constraint(NB,\B)]) ? 
        7      4 Call: solve([constraint(0,1/\NB),constraint(NB,\B)]) ? 
        8      5 Call: simplify([constraint(0,1/\NB),constraint(NB,\B)],_9451) ? s
        8      5 Exit: simplify([constraint(0,1/\0),constraint(0,\1)],[]) ? 
        9      5 Call: solve([]) ? 
        9      5 Exit: solve([]) ? 
        7      4 Exit: solve([constraint(0,1/\0),constraint(0,\1)]) ? 
        5      3 Exit: solve([constraint(0,1/\0),constraint(1,\0),constraint(0,\1)]) ? 
        3      2 Exit: solve([constraint(1,1/\1),constraint(1,\0),constraint(0,1/\0),constraint(1,\0),constraint(0,\1),constraint(1,\0)]) ? 
        1      1 Exit: sat(\ (\0/\ \1)/\ \0) ? 
A = 0,
B = 1 ? ;
no
% trace,source_info
| ?- 
